use std::collections::HashSet;

use crate::guess::*;
use crate::constraints;
use constraints::Constraints;

pub trait Evaluator<const N: usize> {
    fn push(&mut self, _result: GuessResult<N>) {}
    fn evaluate(&self, word: Word<N>) -> Matches<N>;
}

pub struct SecretEvaluator<const N: usize> {
    secret: Word<N>,
}

/// Evaluates guesses according to the provided secret word.
impl<const N: usize> SecretEvaluator<N> {
    pub fn new(secret: Word<N>) -> Self {
        Self {
            secret
        }
    }
}

impl<const N: usize> Evaluator<N> for SecretEvaluator<N> {
    fn evaluate(&self, word: Word<N>) -> Matches<N> {
        guess(&word, &self.secret)
    }
}


/// Evaluates guesses "adversarially", that is, in a way that maximizes the
/// remaining search space after each guess.
pub struct AdversarialEvaluator<const N: usize> {
    possible_words: Dictionary<N>,
    constraints: constraints::Total<N>,
}

impl<const N: usize> AdversarialEvaluator<N> {
    pub fn new(dictionary: &Dictionary<N>) -> Self {
        Self {
            possible_words: dictionary.to_vec(),
            constraints: constraints::Total::new(),
        }
    }

    pub fn possible_word(&self) -> Option<&Word<N>> {
        self.possible_words.first()
    }

    pub fn best_secrets(&self, word: Word<N>) -> Vec<Word<N>> {
        let mut evaluated_matches: HashSet<Matches<N>> = HashSet::new();
        let mut best_secrets: Vec<Word<N>> = Vec::new();
        let mut best_possible_words_count: usize = 0;
        let mut resulting_constraints = constraints::Total::new();

        for possible_secret in &self.possible_words {
            let matches = guess(&word, &possible_secret);
            if !evaluated_matches.insert(matches) { continue; }
            
            resulting_constraints.clone_from(&self.constraints);
            resulting_constraints.push(GuessResult { word, matches });
            
            let possible_words_count = if *possible_secret == word {
                0 // Special case to avoid giving the win when 2 possible words are left
            } else {
                self.possible_words
                    .iter()
                    .filter(|&word| resulting_constraints.check(*word))
                    .count()
            };

            if possible_words_count > best_possible_words_count {
                best_possible_words_count = possible_words_count;
                best_secrets.clear();
            }

            if possible_words_count == best_possible_words_count {
                best_secrets.push(*possible_secret);
            }
        }

        best_secrets
    }
}

impl<const N: usize> Evaluator<N> for AdversarialEvaluator<N> {
    fn evaluate(&self, word: Word<N>) -> Matches<N> {
        let best_secrets = self.best_secrets(word);
        return guess(&word, best_secrets.first().unwrap());
    }

    fn push(&mut self, result: GuessResult<N>) {
        self.constraints.push(result);
        let constraints = &self.constraints;
        self.possible_words.retain(|&word| constraints.check(word));
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use std::convert::TryInto;

    #[test]
    fn it_evaluates_adversarially() {
        let dictionary: Vec<Word<2>> = vec![
            word!("am"), word!("an"), word!("as"), word!("at"), word!("ax"),
                         word!("in"), word!("is"), word!("it"),
                         word!("on"),                           word!("ox"),
        ];

        let adversary = AdversarialEvaluator::new(&dictionary);

        // Locking in the 'a' leaves 4 possibilities;
        // excluding it would leave 2 with the 'n' or 3 without
        assert_eq!(
            [LetterMatch::HERE, LetterMatch::NOWHERE],
            adversary.evaluate(word!("an")),
        );

        // The 'm' doesn't eliminate any non-'a' words,
        // so excluding the 'a' leaves all 5 of them
        assert_eq!(
            [LetterMatch::NOWHERE, LetterMatch::NOWHERE],
            adversary.evaluate(word!("am")),
        );
    }

    #[test]
    fn it_adheres_to_constraints() {
        let dictionary: Vec<Word<2>> = vec![
            word!("am"), word!("an"), word!("as"), word!("at"), word!("ax"),
                         word!("in"), word!("is"), word!("it"),
                         word!("on"),                           word!("ox"),
        ];

        let mut adversary = AdversarialEvaluator::new(&dictionary);
        adversary.push(GuessResult {
            word: word!("an"),
            matches: [LetterMatch::NOWHERE, LetterMatch::NOWHERE],
        });

        // We have already excluded both of these letters
        assert_eq!(
            [LetterMatch::NOWHERE, LetterMatch::NOWHERE],
            adversary.evaluate(word!("an")),
        );

        // Locking in the 'i' leaves 2 possibilities;
        // excluding it would only leave 'ox'
        assert_eq!(
            [LetterMatch::HERE, LetterMatch::NOWHERE],
            adversary.evaluate(word!("in")),
        );

        // Play it out
        adversary.push(GuessResult {
            word: word!("in"),
            matches: [LetterMatch::HERE, LetterMatch::NOWHERE],
        });

        // Leaves 'it'
        assert_eq!(
            [LetterMatch::HERE, LetterMatch::NOWHERE],
            adversary.evaluate(word!("is")),
        );

        adversary.push(GuessResult {
            word: word!("is"),
            matches: [LetterMatch::HERE, LetterMatch::NOWHERE],
        });

        // No choice but to accept
        assert_eq!(
            [LetterMatch::HERE, LetterMatch::HERE],
            adversary.evaluate(word!("it")),
        );
    }
}