#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
pub enum LetterMatch {
    NOWHERE,
    ELSEWHERE,
    HERE,
}

pub type Word<const N: usize> = [char; N];
pub type Matches<const N: usize> = [LetterMatch; N];

pub fn guess<const N: usize>(word: &Word<N>, secret: &Word<N>) -> Matches<N> {
    let mut result = [LetterMatch::NOWHERE; N];
    let mut matched_secret_letters = [false; N];

    // Find exact matches
    for a in 0..N {
        if word[a] == secret[a] {
            result[a] = LetterMatch::HERE;
            matched_secret_letters[a] = true;
        }
    }

    // Find wrong-position matches
    for a in 0..N {
        for b in 0..N {
            if a == b || result[a] != LetterMatch::NOWHERE || matched_secret_letters[b] {
                continue;
            }
            if word[a] == secret[b] {
                result[a] = LetterMatch::ELSEWHERE;
                matched_secret_letters[b] = true;
                break;
            }
        }
    }

    result
}

#[derive(Copy, Clone, Debug)]
pub struct GuessResult<const N: usize> {
    pub word: Word<N>,
    pub matches: Matches<N>,
}

impl<const N: usize> GuessResult<N> {
    pub fn correct(&self) -> bool {
        for a in 0..N {
            if self.matches[a] != LetterMatch::HERE {
                return false;
            }
        }
        true
    }
}

impl<const N: usize> std::fmt::Display for GuessResult<N> {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        let mut result = Vec::new();
        for a in 0..N {
            result.push(match self.matches[a] {
                LetterMatch::ELSEWHERE => format!("({})", self.word[a]),
                LetterMatch::HERE => format!("[{}]", self.word[a]),
                LetterMatch::NOWHERE => format!(" {} ", self.word[a]),
            })
        }
        write!(f, "{}", result.join(""))
    }
}

pub type GuessHistory<const N: usize> = Vec<GuessResult<N>>;
pub type Dictionary<const N: usize> = Vec<Word<N>>;


#[cfg(test)]
mod tests {
    use super::*;
    use std::convert::TryInto;

    #[test]
    fn it_matches_correct_guess() {
        let secret: Word<5> = word!("loose");
        assert_eq!([LetterMatch::HERE; 5], guess(&secret, &secret));
    }
    
    #[test]
    fn it_matches_all_elsewhere() {
        let secret = word!("loose");
        let word = word!("seloo"); // lol
        assert_eq!([LetterMatch::ELSEWHERE; 5], guess(&word, &secret));

    }

    #[test]
    fn it_matches_all_nowhere() {
        let secret = word!("loose");
        let word = word!("brink");
        assert_eq!([LetterMatch::NOWHERE; 5], guess(&word, &secret));
    }

    #[test]
    fn it_matches_close_guess() {
        let secret = word!("loose");
        let word = word!("loves");
        let expected = [LetterMatch::HERE, LetterMatch::HERE, LetterMatch::NOWHERE, LetterMatch::ELSEWHERE, LetterMatch::ELSEWHERE];
        assert_eq!(expected, guess(&word, &secret));
    }
}
